# Measuring Thrust and Drag Constants[^1]

## Introduction

For proper modeling of the quadrotor we need to go through the identification process.
This involves measuring parameters of the specific quadrotor in use, including both the thrust and drag constants.
This document describes the setup and procedure for measuring the thrust and drag constants of a quadrotor.

## Setup 

The materials required are the following:
* Function generator
* Power supply capable of outputting ~30A
* Digital scale
* Optical tachometer
* Quadrotor

The first thing needed is to set the quadrotor onto the scale facing upside down.
This directs the force generated by the quad directly onto the scale
(note that you may need to place an object in between the quad and the scale to eliminate the possibility of a rotor hitting the scale).
After this is done the scale needs to be zeroed out so that the only force being measured is produced by the quads thrust.

Next we need to power the quad with the power supply.
This will replace the LiPO battery so that we can accurately measure the voltage being supplied at all times.
The voltage from the power supply should be ~12V.

The next step is to connect the ESC[^2] signal lines to the function generator to produce a known PWM output.
This PWM output has the following characteristics:
* $`f`$ = 400 Hz Square Wave
* $`V_{pp}`$ = 3.2 V
* $`V_{\text{offset}}`$ = 1.6 V
* Needs a duty cycle of 40% (1 millisecond pulse width) to initialize the ESCs

In the case that the onboard ESCs have changed,
the frequency and duty cycle required for initialization may change.
Simply reference the ESC documentation for the compatible signal frequencies
and choose a frequency within that range.
From there the calculation of the duty cycle for initialization,
denoted by $`P_{\text{min}}`$,  is as follows:

```math
P_{\text{min}} = 0.001f
```

To calculate the maximum duty cycle that the ESC can handle, denoted by $`P_{\text{max}}`$, we use the following equation:

```math
P_{\text{max}} = 0.002f
```

These ESC signal lines are connected to a ribbon cable,
which then connects to the zybo board
through an insulation-displacement connector (IDC).
The function generator inputs go directly to the ESC
through this IDC connector and ribbon cable.
The pinout of the IDC connector is as follows:

| Pin 1[^3] | Pin 3 | Pin 5 | Pin 7 | Pin  9 | Pin 11 |
|:---------:|:-----:|:-----:|:-----:|:------:|:------:|
| Pin 2     | Pin 4 | Pin 6 | Pin 8 | Pin 10 | Pin 12 |

| Pin Number | Purpose                |
|:----------:|:-----------------------|
| Pin 1      | Throttle               |
| Pin 2      | ESC 0 signal line      |
| Pin 3      | Aileron                |
| Pin 4      | ESC 1 signal line      |
| Pin 5      | Elevator               |
| Pin 6      | ESC 2 signal line      |
| Pin 7      | Rudder                 |
| Pin 8      | ESC 3 signal line      |
| Pin 9      | Ground RC Receiver     |
| Pin 10     | Ground                 |
| Pin 11     | Not used               |
| Pin 12     | $`V_{cc}`$ RC Receiver |

After the ESC signal lines are properly connected
and the ESCs are being properly powered by the power supply,
we can begin collecting data.

## Data Collection

The first thing we need to do is increase the duty cycle of the function generator
from the calculated initialization value until the motors begin to turn.
We will denote this duty cycle value as $`P_i`$ for reference.
If the rotors do not begin to turn,
verify that the wave being sent from the function generator is correct
and that the ESCs are being powered properly from the power supply.

The data collection process is as follows:

1. Increase the input duty cycle by 1% (starting at $`P_i`$) and record its current value.
1. Using the optical tachometer, record the approximate steady state speed reached by each motor/rotor in RPM.
1. Record the corresponding input voltage from the power supply.
1. Record the corresponding digital scale reading in grams.
1. Record the current draw from the power supply.[^4]
1. Repeat 1-5 until reaching $`P_{\text{max}}`$.

To utilize the later provided MATLAB scripts to analyze the collected data and calculate the thrust and drag constants, the data needs to be stored in a specific way. Firstly, Excel should be used for storing the data. The required header layout of this Excel file should be as follows:


|Duty Cycle|Rotor 0 RPM|Rotor 1 RPM|Rotor 2 RPM|Rotor 3 RPM|Voltage|Scale Reading|Current|
|:--------:|:---------:|:---------:|:---------:|:---------:|:-----:|:-----------:|:-----:|

This is due to the fact that the functions created for calculating
the thrust and drag constants take in data as a table data type in MATLAB.
How these functions access the data in this table
is through specific indexing shown below

|Index 1|Index 2|Index 3|Index 4|Index 5|Index 6|Index 7|Index 8|
|:-----:|:-----:|:-----:|:-----:|:-----:|:-----:|:-----:|:-----:|

where each index value corresponds to its respective header.
Therefore, the name of the header doesn’t matter
as long as its associated data is in the same column as its referenced index.

## MATLAB Functions and Script

### Calculate Thrust Constant Function

```matlab
function [ Kt ] = calc_thrust_constant( data )
%CALC_THRUST_CONSTANT
%   Calculates the thrust constant (Kt) given experimental data. The thrust
%   constant is described in detail in sections 4.2.1.1 and 5.5.1 of 
%   "Model development, system identification, and control of a quadrotor
%   helicopter" by Matt Rich.

%   Input Arguments:
%   data: experimental data

% Convert RPM to angular speed of each rotor.
rotor_speed_0 = data.(2) * (pi/30);
rotor_speed_1 = data.(3) * (pi/30);
rotor_speed_2 = data.(4) * (pi/30);
rotor_speed_3 = data.(5) * (pi/30);

% Define the A matrix as the sum of each rotors speed squared.
A = (rotor_speed_0.^2 + rotor_speed_1.^2 + rotor_speed_2.^2 + rotor_speed_3.^2);

% Convert weight (g) to thrust force (N) by converting grams to kilograms and
% multiplying by the acceleration of gravity.
T = (data.(7)/1000)*9.8;

% Calculate the thrust constant (Kt) through the following least squares
% approximation: Kt = (A'A)^-1.*A'.*T as defined on page 65 of "Model
% development, system identification, and control of a quadrotor
% helicopter" (Matt Rich's Thesis).
Kt = ((A'*A)^(-1))*(A'*T) 

end
```

### Calculate Drag Constant Function 

```matlab
function [ Kd ] = calc_drag_constant( data, Pmin, Pmax, Rm, Kv, Kq, If )
%CALC_DRAG_CONSTANT 
%   Calculates the drag constant (Kd) given experimental data. The drag
%   constant is described in detail in sections 4.2.5, 4.2.6, and 5.5.4.1 
%   of "Model development, system identification, and control of a 
%   quadrotor helicopter" by Matt Rich.

%   Input Arguments:
%   data: experimental data
%   Pmax: Calculated maximum duty cycle of the function generators PWM wave that the ESC can  %   handle.
%   Pmin: Calculated minimum duty cycle of the function generators PWM wave for initialization %   of the ESC.
%   Rm: Motor resistance
%   Kv: Back-EMF constant of the motor
%   Kq: Torque constant of the motor
%   If: No-Load (friction) current of the motor


% Convert RPM to angular speed of each rotor.
rotor_speed_0 = data.(2) * (pi/30);
rotor_speed_1 = data.(3) * (pi/30);
rotor_speed_2 = data.(4) * (pi/30);
rotor_speed_3 = data.(5) * (pi/30);

% Refer to the sections described in the header of this function
% for a better understanding of what this loop is doing.
% u: Defined in section 4.2.5
% omega: vector of each rotors speed in rad/s
% A: column vector of each rotors speed squared
% b: Defined in section 5.5.4.1
% Kd_vector: Vector containing all experimental Kd values
for i = 1:length(rotor_speed_1)
    u = ((data.(1)(i)/100) - Pmin)/(Pmax - Pmin);
    omega = [rotor_speed_0(i), rotor_speed_1(i), rotor_speed_2(i), rotor_speed_3(i)];
    A = omega'.^2;
    b = ((u*data.(6)(i))/(Rm*Kq))-omega'/(Rm*Kq*Kv)-If/Kq;
    Kd_vector(i) = ((A'*A)^(-1))*(A'*b);
end

% Calculate the average of the Kd_vector to obtain the drag constant.
Kd = sum(Kd_vector)/length(Kd_vector)

end
```

Note that the function for calculating the drag constant
is dependant upon the following motor parameters:

| Parameter | Units                | Description                |
|:---------:|:--------------------:|:---------------------------|
| $`R_m`$   | $`\Omega`$           | Motor Resistance           |
| $`K_V`$   | $`\frac{rad}{V\,s}`$ | Back-emf constant          |
| $`K_Q`$   | $`\frac{A}{N\,m}`$   | Torque constant            |
| $`i_f`$   | $`A`$                | No-Load (friction) current |

### Script for Calculating both Thrust and Drag Constants

```matlab
% Import data as a table.
data = readtable(‘file location’);

% Define Pmin and Pmax, as well as the motor parameters Rm, Kv, Kq, and If
Pmin = ‘insert calculated Pmin value here’;
Pmax = ‘insert calculated Pmax value here’;
Rm = ‘insert Rm value here’;
Kv = ‘insert Kv value here’;
Kq = ‘insert Kq value here’;
If = ‘insert If value here’;

% Call the calc_thrust_constant() function.
Kt = calc_thrust_constant(data);

% Call the calc_drag_constant() function.
Kd = calc_drag_constant(data, Pmin, Pmax, Rm, Kv, Kq, If);
```

[^1]: This document was written in October 2016 by Andy Snawerdt, Tara Mina, and Brendan Bartels.

[^2]: This document's specifications references the current ESCs onboard, which are the [DJI 30A opto](https://www.dji.com/en/index.php/ESC).

[^3]: Pin 1 is specified by a trinagle on the body of the IDC connector.

[^4]: This isn’t actually necessary for the thrust or drag constant calculations, however this information may be useful later on for power requirements of the rotors.
