#include "mainwindow.h"
#include "ui_mainwindow.h"

#include <QFileDialog>
#include <QThread>
#include <QTimer>
#include <QRegExp>
#include <QProcessEnvironment>
#include <QPixmap>
#include <QProcess>

#include "trackerworker.h"
#include "controlworker.h"
#include "graph_blocks.h"
#include "quaditem.h"
#include "slotprocess.h"
#include "qFlightInstruments.h"

#include <iostream>
MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    setpointList(new QStandardItemModel(this)),
    nextSpTimer(new QTimer(this)),
    sp_x(0.0f),
    sp_y(0.0f),
    sp_z(0.0f),
    trackerTimer(new QTimer(this)),
    workerStartTimer(new QTimer(this)),
    backendProcess(new QProcess(this)),
    connectedWorkers(0)
{
    ui->setupUi(this);

    QGraphicsScene *posScene = new QGraphicsScene(this);
    ui->posView->setScene(posScene);
    QuadItem * posIndicator = new QuadItem();
    posScene->addItem(posIndicator);

    QGraphicsScene *attScene = new QGraphicsScene(this);
    ui->attView->setScene(attScene);
    QADI * attIndicator = new QADI();
    attScene->addWidget(attIndicator);


    workerStartTimer->setSingleShot(true);

    /* Create a thread for workers */
    QThread* workerThread = new QThread(this);

    /* Create a worker to update the tracker */
    TrackerWorker * trackerWorker = new TrackerWorker();

    /* Move it to the worker thread. This means that slots of this worker
     * will run in the worker thread, and not block the UI */
    trackerWorker->moveToThread(workerThread);

    /* Connect tracker worker */
    connect(trackerWorker, SIGNAL (finished(float, float, float, float, float, float)),
            this, SLOT (updateTracker(float, float, float, float, float, float)));

    connect(trackerWorker, SIGNAL (finished(float, float, float, float, float, float)),
            posIndicator, SLOT(updatePos(float,float,float,float,float,float)));

    connect(trackerWorker, SIGNAL (finished(float, float, float, float, float, float)),
            attIndicator, SLOT(updateAttitude(float, float, float, float, float, float)));

    backendProcess->setProcessChannelMode(QProcess::MergedChannels);

    /* Create another worker for the control graph */
    QThread * cwThread = new QThread(this);
    ControlWorker * controlWorker = new ControlWorker();
    controlWorker->moveToThread(cwThread);

    /* Connect signals from control worker */
    connect(controlWorker, SIGNAL (gotNodes(QStringList)), this, SLOT (newNodes(QStringList)));
    connect(controlWorker, SIGNAL (gotParams(QStringList)), this, SLOT (newParams(QStringList)));
    connect(controlWorker, SIGNAL (gotParamValue(QString, QString, float)), this, SLOT (newParamValue(QString, QString, float)));
    connect(controlWorker, SIGNAL (gotNodeOutput(QString, float)), this, SLOT (newNodeOutput(QString, float)));
    connect(controlWorker, SIGNAL (gotConstantBlocks(QStringList)), this, SLOT (newConstantBlocks(QStringList)));
    connect(controlWorker, SIGNAL (graphRendered(QString)), this, SLOT (newControlGraph(QString)));
    connect(controlWorker, SIGNAL (paramSet(QString, QString)), controlWorker, SLOT (getParamValue(QString, QString)));

    /* Signals to control worker */
    connect(ui->pbControlRefresh, SIGNAL (clicked()), controlWorker, SLOT (getNodes()));
    connect(ui->nodeSelect, SIGNAL (currentIndexChanged(QString)), controlWorker, SLOT (getParams(QString)));
    connect(this, SIGNAL (getParamValue(QString, QString)), controlWorker, SLOT (getParamValue(QString, QString)));
    connect(this, SIGNAL (setParamValue(QString, QString, float)), controlWorker, SLOT (setParamValue(QString, QString, float)));
    connect(this, SIGNAL (getNodeOutput(QString)), controlWorker, SLOT (getNodeOutput(QString)));

    /* Connect and disconnect from backend when signals emitted */
    connect(workerStartTimer, SIGNAL (timeout()), trackerWorker, SLOT (connectBackend()));
    connect(workerStartTimer, SIGNAL (timeout()), controlWorker, SLOT (connectBackend()));
    connect(this, SIGNAL (connectWorkers()), trackerWorker, SLOT (connectBackend()));
    connect(this, SIGNAL (connectWorkers()), controlWorker, SLOT (connectBackend()));
    connect(this, SIGNAL (disconnectWorkers()), trackerWorker, SLOT (disconnectBackend()));
    connect(this, SIGNAL (disconnectWorkers()), controlWorker, SLOT (disconnectBackend()));

    connect(backendProcess, SIGNAL (started()), this, SLOT (backendStarted()));
    connect(backendProcess, SIGNAL (errorOccurred(QProcess::ProcessError)), this, SLOT (backendError(QProcess::ProcessError)));
    connect(backendProcess, SIGNAL (finished(int, QProcess::ExitStatus)), this, SLOT (backendFinished(int, QProcess::ExitStatus)));
    connect(backendProcess, SIGNAL (readyReadStandardOutput()), this, SLOT (backendRead()));
    connect(backendProcess, SIGNAL (readyReadStandardError()), this, SLOT (backendRead()));

    connect(controlWorker, SIGNAL (connected()), this, SLOT (workerConnected()));
    connect(controlWorker, SIGNAL (disconnected()), this, SLOT (workerDisconnected()));

    connect(trackerWorker, SIGNAL (connected()), this, SLOT (workerConnected()));
    connect(trackerWorker, SIGNAL (disconnected()), this, SLOT (workerDisconnected()));

    /* Connect refresh button and refresh timer to tracker worker */
    connect(trackerTimer, SIGNAL(timeout()), this, SLOT(updatePosAtt()));
    connect(ui->pbRefresh, SIGNAL (clicked()), this, SLOT (updatePosAtt()));

    connect(this, SIGNAL(getPosAttFromBackend()), trackerWorker, SLOT(process()));

    /* Timer used for next setpoint */
    nextSpTimer->setSingleShot(true);
    connect(nextSpTimer, SIGNAL (timeout()), this, SLOT (on_pbNextSetpoint_clicked()));

    /* Start the things */
    trackerTimer->start(100);
    workerThread->start();
    cwThread->start();

    /* Connect the setpointlist to the model */
    ui->setpointList->setModel(setpointList);

    /* Connect various things that can result in sending setpoints */
    connect(ui->pbSendSetpoint, SIGNAL (clicked()), this, SLOT (sendSetpoints()));
    connect(ui->setpointList, SIGNAL (doubleClicked(QModelIndex)), this, SLOT (sendSelectedSetpoint()));

    /* Populate scripts list */
    QDir scriptsDir("scripts/");
    QStringList scripts = scriptsDir.entryList();
    for (int i = 0; i < scripts.size(); i++) {
        QAction * action = ui->menuScripts->addAction(scripts[i]);
        SlotProcess * scriptProcess = new SlotProcess(action);
        scriptProcess->setProgram(scriptsDir.filePath(scripts[i]));
        connect(action, SIGNAL (triggered()), scriptProcess, SLOT (startProcess()));
    }
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::updatePosAtt()
{
    if (ui->posattSrcVrpn->isChecked()) {
        emit(getPosAttFromBackend());
    } else if (ui->posattSrcQuad->isChecked()) {
        emit(getNodeOutput(ui->xPositionSelect->currentText()));
        emit(getNodeOutput(ui->yPositionSelect->currentText()));
        emit(getNodeOutput(ui->zPositionSelect->currentText()));
        emit(getNodeOutput(ui->pAttitudeSelect->currentText()));
        emit(getNodeOutput(ui->rAttitudeSelect->currentText()));
        emit(getNodeOutput(ui->yAttitudeSelect->currentText()));
    }
}

void MainWindow::updateTracker(float x, float y, float z, float p, float r, float yaw)
{
    ui->xActual->setText(QString::number(x));
    ui->yActual->setText(QString::number(y));
    ui->zActual->setText(QString::number(z));
    ui->pitchActual->setText(QString::number(p));
    ui->rollActual->setText(QString::number(r));
    ui->yawActual->setText(QString::number(yaw));

    float dist = sqrt(pow(x - sp_x, 2.0) + pow(y - sp_y, 2.0) + pow(z - sp_z, 2.0));

    ui->dist->setText(QString::number(dist));

    if (!nextSpTimer->isActive() && ui->autonavEnabled->isChecked() &&
            (dist < ui->autonavThreshold->text().toFloat())) {
        nextSpTimer->start(ui->autonavDelay->text().toInt());
    }
}

void MainWindow::on_pbStart_clicked()
{
    backendProcess->setProgram(ui->backendPath->text());

    backendProcess->start();
}

void MainWindow::backendStarted()
{
    ui->pbStart->setEnabled(false);
    ui->pbStop->setEnabled(true);
    workerStartTimer->start(750);
}

void MainWindow::backendError(QProcess::ProcessError error)
{
    ui->vConsole->append(backendProcess->errorString());
}

void MainWindow::backendRead()
{
    ui->vConsole->append(backendProcess->readAll());
}

void MainWindow::workerConnected()
{
    connectedWorkers++;
     if (connectedWorkers == 2) {
        ui->pbStart->setEnabled(false);
        ui->pbConnect->setEnabled(false);
        ui->pbStop->setEnabled(true);
    }
}

void MainWindow::workerDisconnected()
{
    connectedWorkers--;
    if (( connectedWorkers == 0) && (backendProcess->state() == QProcess::Running)) {
        backendProcess->terminate();
    } else if (connectedWorkers == 0) {
        ui->pbStart->setEnabled(true);
        ui->pbConnect->setEnabled(true);
        ui->pbStop->setEnabled(false);
    }
}

void MainWindow::backendFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
    ui->pbStart->setEnabled(true);
    ui->pbConnect->setEnabled(true);
    ui->pbStop->setEnabled(false);
}

void MainWindow::on_pbConnect_clicked()
{
    emit(connectWorkers());
}

void MainWindow::on_pbStop_clicked()
{
    emit(disconnectWorkers());
}

void MainWindow::on_chooseBackend_clicked()
{
    QString backendPath = QFileDialog::getOpenFileName(this,
         tr("Path to Backend Executable"));
    ui->backendPath->setText(backendPath);
}


void MainWindow::newNodes(QStringList blocks)
{
    QComboBox * select = ui->nodeSelect;
    select->clear();
    select->addItems(blocks);

    this->ui->noGraphWarning1->setVisible(false);
    this->ui->noGraphWarning2->setVisible(false);
    this->ui->noGraphWarningLine->setVisible(false);
}

void MainWindow::newConstantBlocks(QStringList blocks)
{
    ui->xSetpointSelect->clear();
    ui->xSetpointSelect->addItems(blocks);

    ui->ySetpointSelect->clear();
    ui->ySetpointSelect->addItems(blocks);

    ui->zSetpointSelect->clear();
    ui->zSetpointSelect->addItems(blocks);

    ui->yawSetpointSelect->clear();
    ui->yawSetpointSelect->addItems(blocks);

    ui->xPositionSelect->clear();
    ui->xPositionSelect->addItems(blocks);

    ui->yPositionSelect->clear();
    ui->yPositionSelect->addItems(blocks);

    ui->zPositionSelect->clear();
    ui->zPositionSelect->addItems(blocks);

    ui->pAttitudeSelect->clear();
    ui->pAttitudeSelect->addItems(blocks);

    ui->rAttitudeSelect->clear();
    ui->rAttitudeSelect->addItems(blocks);

    ui->yAttitudeSelect->clear();
    ui->yAttitudeSelect->addItems(blocks);

    for (ssize_t i = 0; i < blocks.size(); i++) {
        if (blocks[i].contains("setpoint", Qt::CaseInsensitive) || blocks[i].contains("sp", Qt::CaseInsensitive)) {
            if (blocks[i].contains("x ", Qt::CaseInsensitive)) {
                ui->xSetpointSelect->setCurrentIndex(i);
            }
            if (blocks[i].contains("y ", Qt::CaseInsensitive)) {
                ui->ySetpointSelect->setCurrentIndex(i);
            }
            if (blocks[i].contains("z ", Qt::CaseInsensitive) || blocks[i].contains("alt", Qt::CaseInsensitive)) {
                ui->zSetpointSelect->setCurrentIndex(i);
            }
            if (blocks[i].contains("yaw", Qt::CaseInsensitive) || blocks[i].contains("alt", Qt::CaseInsensitive)) {
                ui->yawSetpointSelect->setCurrentIndex(i);
            }
        }

        if (blocks[i] == QString("Pitch")) {
            ui->pAttitudeSelect->setCurrentIndex(i);
        }

        if (blocks[i] == QString("Roll")) {
            ui->rAttitudeSelect->setCurrentIndex(i);
        }

        if (blocks[i] == QString("Yaw")) {
            ui->yAttitudeSelect->setCurrentIndex(i);
        }

        if (blocks[i] == QString("Lidar")) {
            ui->zPositionSelect->setCurrentIndex(i);
        }

        if (blocks[i] == QString("OF X Trim Add")) {
            ui->xPositionSelect->setCurrentIndex(i);
        }

        if (blocks[i] == QString("OF Y Trim Add")) {
            ui->yPositionSelect->setCurrentIndex(i);
        }
    }
}

void MainWindow::newNodeOutput(QString node, float val)
{
    /* Update the nav page if quad is set as the source for pos/att */
    if (node == ui->xPositionSelect->currentText()) {
        ui->xActual->setText(QString::number(val));
    } else if (node == ui->yPositionSelect->currentText()) {
        ui->yActual->setText(QString::number(val));
    } else if (node == ui->zPositionSelect->currentText()) {
        ui->zActual->setText(QString::number(val));
    } else if (node == ui->pAttitudeSelect->currentText()) {
        ui->pitchActual->setText(QString::number(val));
    } else if (node == ui->rAttitudeSelect->currentText()) {
        ui->rollActual->setText(QString::number(val));
    } else if (node == ui->yAttitudeSelect->currentText()) {
        ui->yawActual->setText(QString::number(val));
    }
}

void MainWindow::newParams(QStringList params)
{
    QComboBox * select = ui->paramSelect;
    select->clear();
    select->addItems(params);
}

void MainWindow::newParamValue(QString node, QString param, float val)
{
    ui->paramValue->setText(QString::number(val));

    /* Update the nav page setpoints if it's a setpoint paramvalue */
    if (node == ui->xSetpointSelect->currentText()) {
        ui->xSetpoint->setText(QString::number(val));
    } else if (node == ui->ySetpointSelect->currentText()) {
        ui->ySetpoint->setText(QString::number(val));
    } else if (node == ui->zSetpointSelect->currentText()) {
        ui->zSetpoint->setText(QString::number(val));
    }
}

void MainWindow::on_paramSelect_currentIndexChanged(const QString &arg1)
{
    emit(getParamValue(ui->nodeSelect->currentText(), arg1));
}

void MainWindow::on_paramValue_returnPressed()
{
    emit (setParamValue(ui->nodeSelect->currentText(),
                        ui->paramSelect->currentText(),
                        ui->paramValue->text().toFloat()));
}

void MainWindow::sendSetpoints()
{
    if (ui->sendX->isChecked()) {
        sp_x = ui->xSetpoint->text().toFloat();
        emit (setParamValue(ui->xSetpointSelect->currentText(),
                            blockDefs[BLOCK_CONSTANT]->param_names[0], sp_x));
    }

    if (ui->sendY->isChecked()) {
        sp_y = ui->ySetpoint->text().toFloat();
        emit (setParamValue(ui->ySetpointSelect->currentText(),
                            blockDefs[BLOCK_CONSTANT]->param_names[0], sp_y));
    }

    if (ui->sendZ->isChecked()) {
        sp_z = ui->zSetpoint->text().toFloat();
        emit (setParamValue(ui->zSetpointSelect->currentText(),
                            blockDefs[BLOCK_CONSTANT]->param_names[0], sp_z));
    }

    if (ui->sendYaw->isChecked()) {
        emit (setParamValue(ui->yawSetpointSelect->currentText(),
                            blockDefs[BLOCK_CONSTANT]->param_names[0],
                            ui->yawSetpoint->text().toFloat()));
    }
}

void MainWindow::on_pbAppendSetpoint_clicked()
{
    QString str("[" + ui->xSetpoint->text() + ", "+
                ui->ySetpoint->text() + ", " +
                ui->zSetpoint->text() + ", " +
            ui->yawSetpoint->text() + "]");

    setpointList->appendRow(new QStandardItem(str));
}

void MainWindow::on_pbNextSetpoint_clicked()
{
    QListView * listView = ui->setpointList;
    if (listView->currentIndex().isValid() && setpointList->index(listView->currentIndex().row() + 1, 0).isValid()) {
        listView->setCurrentIndex(setpointList->index(listView->currentIndex().row() + 1, 0));
    } else {
        listView->setCurrentIndex(setpointList->index(0, 0));
    }
    sendSelectedSetpoint();
}


void MainWindow::sendSelectedSetpoint()
{
    if (ui->setpointList->currentIndex().isValid()) {
        QRegExp regex("\\[(.*), (.*), (.*), (.*)\\]");
        int row = ui->setpointList->currentIndex().row();

        regex.indexIn(setpointList->item(row)->text());
        ui->xSetpoint->setText(regex.cap(1));
        ui->ySetpoint->setText(regex.cap(2));
        ui->zSetpoint->setText(regex.cap(3));
        ui->yawSetpoint->setText(regex.cap(4));

        sendSetpoints();
    }
}

void MainWindow::on_pbActualToSetpoint_clicked()
{
    ui->xSetpoint->setText(ui->xActual->text());
    ui->ySetpoint->setText(ui->yActual->text());
    ui->zSetpoint->setText(ui->zActual->text());
    ui->yawSetpoint->setText(ui->yawActual->text());
}

void MainWindow::on_pbDeleteSetpoint_clicked()
{
    if (ui->setpointList->currentIndex().isValid()) {
        setpointList->removeRow(ui->setpointList->currentIndex().row());
    }
}

void MainWindow::newControlGraph(QString graph)
{
    ui->graphImage->setPixmap(QPixmap(graph));
}

void MainWindow::on_pbActualToWaypoint_clicked()
{
    QString str("[" + ui->xActual->text() + ", "+
                ui->yActual->text() + ", " +
                ui->zActual->text() + ", " +
                ui->yawActual->text() + "]");

    setpointList->appendRow(new QStandardItem(str));
}

void MainWindow::on_pbMoveUp_clicked()
{
    if (ui->setpointList->currentIndex().isValid()) {
        int current = ui->setpointList->currentIndex().row();
        if (current > 0) {
            setpointList->insertRow(current - 1, setpointList->takeItem(current));
            setpointList->removeRow(current + 1);
        }
    }
}

void MainWindow::on_pbMoveDown_clicked()
{
    if (ui->setpointList->currentIndex().isValid()) {
        int current = ui->setpointList->currentIndex().row();
        if (current < (setpointList->rowCount() - 1)) {
            setpointList->insertRow(current + 2, setpointList->takeItem(current));
            setpointList->removeRow(current);
        }
    }
}

void MainWindow::on_pbInsertSetpoint_clicked()
{
    int current = 0;
    if (ui->setpointList->currentIndex().isValid()) {
        current = ui->setpointList->currentIndex().row();
    }

    QString str("[" + ui->xSetpoint->text() + ", "+
                ui->ySetpoint->text() + ", " +
                ui->zSetpoint->text() + ", " +
                ui->yawSetpoint->text() + "]");
    setpointList->insertRow(current, new QStandardItem(str));
}

void MainWindow::on_pbSaveWaypoints_clicked()
{
    QString filter = "Waypoints (*.wpt)";
    QString savePath = QFileDialog::getSaveFileName(this, tr("Save waypoint"), QString(), filter, &filter, QFileDialog::DontUseNativeDialog);

    if (!savePath.isEmpty()) {
        QFileInfo finfo(savePath);
        if (finfo.suffix().isNull())
            savePath.append(".wpt");
        QFile f(savePath);
        f.open(QIODevice::WriteOnly | QIODevice::Text);
        for (int i= 0; i < setpointList->rowCount(); i++) {
            f.write(setpointList->item(i)->text().toStdString().c_str());
            f.write("\n");
        }
        f.close();
    }
}

void MainWindow::on_pbLoadWaypoints_clicked()
{

    QString filter = "Waypoints (*.wpt)";
    QString openPath = QFileDialog::getOpenFileName(this, tr("Open Waypoint File"), QString(),filter);

    if (!openPath.isEmpty()) {
        QFile f(openPath);
        f.open(QIODevice::ReadOnly | QIODevice::Text);

        setpointList->clear();
        while (!f.atEnd()) {
            QString line = f.readLine();
            line.chop(1);
            setpointList->appendRow(new QStandardItem(line));
        }
        f.close();
    }
}

void MainWindow::on_xSetpoint_returnPressed()
{
    sp_x = ui->xSetpoint->text().toFloat();
    emit (setParamValue(ui->xSetpointSelect->currentText(),
                        blockDefs[BLOCK_CONSTANT]->param_names[0], sp_x));
}

void MainWindow::on_ySetpoint_returnPressed()
{
    sp_y = ui->ySetpoint->text().toFloat();
    emit (setParamValue(ui->ySetpointSelect->currentText(),
                        blockDefs[BLOCK_CONSTANT]->param_names[0], sp_y));
}

void MainWindow::on_zSetpoint_returnPressed()
{
    sp_z = ui->zSetpoint->text().toFloat();
    emit (setParamValue(ui->zSetpointSelect->currentText(),
                        blockDefs[BLOCK_CONSTANT]->param_names[0], sp_z));
}

void MainWindow::on_yawSetpoint_returnPressed()
{
    emit (setParamValue(ui->yawSetpointSelect->currentText(),
                        blockDefs[BLOCK_CONSTANT]->param_names[0],
            ui->yawSetpoint->text().toFloat()));
}

void MainWindow::on_posattSrcVrpn_clicked()
{
    trackerTimer->setInterval(100);
}

void MainWindow::on_posattSrcQuad_clicked()
{
    trackerTimer->setInterval(750);
}
